/*!\file ControlInput.c
 * \brief: implementation of the ControlInput object
 */

#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#error "Cannot compile with HAVE_CONFIG_H symbol! run configure first!"
#endif

#include "stdio.h"
#include <string.h>
#include "../objects.h"
#include "../../EnumDefinitions/EnumDefinitions.h"
#include "../../shared/shared.h"
#include "../../Container/Container.h"
#include "../../include/include.h"

/*ControlInput constructors and destructor*/
/*FUNCTION ControlInput::ControlInput(){{{1*/
ControlInput::ControlInput(){
	values     =NULL;
	savedvalues =NULL;
	gradient   =NULL;
}
/*}}}*/
/*FUNCTION ControlInput::ControlInput(int enum_type,int enum_input,double* pvalues){{{1*/
ControlInput::ControlInput(int in_enum_type,int enum_input,double* pvalues){
	enum_type=in_enum_type;
	switch(enum_input){
		case TriaVertexInputEnum:
			values     =new TriaVertexInput(enum_type,pvalues);
			savedvalues=new TriaVertexInput(enum_type,pvalues);
			break;
		case PentaVertexInputEnum:
			values     =new PentaVertexInput(enum_type,pvalues);
			savedvalues=new PentaVertexInput(enum_type,pvalues);
			break;
		default:
			ISSMERROR("Input of Enum %s not supported yet by ControlInput",EnumToString(enum_input));
	}
	gradient   =NULL;
}
/*}}}*/
/*FUNCTION ControlInput::~ControlInput(){{{1*/
ControlInput::~ControlInput(){
	delete values;
	delete savedvalues;
	delete gradient;
}
/*}}}*/

/*Object virtual functions definitions:*/
		/*FUNCTION ControlInput::Echo {{{1*/
void ControlInput::Echo(void){
	this->DeepEcho();
}
/*}}}*/
/*FUNCTION ControlInput::DeepEcho{{{1*/
void ControlInput::DeepEcho(void){

	printf("ControlInput:\n");
	printf("   enum: %i (%s)\n",this->enum_type,EnumToString(this->enum_type));
	printf("---values: \n");      if (values)      values->Echo();
	printf("---savedvalues: \n"); if (savedvalues) savedvalues->Echo();
	printf("---gradient: \n");    if (gradient)    gradient->Echo();
}
/*}}}*/
/*FUNCTION ControlInput::Id{{{1*/
int    ControlInput::Id(void){ return -1; }
/*}}}*/
/*FUNCTION ControlInput::MyRank{{{1*/
int    ControlInput::MyRank(void){ 
	extern int my_rank;
	return my_rank; 
}
/*}}}*/
/*FUNCTION ControlInput::Marshall{{{1*/
void  ControlInput::Marshall(char** pmarshalled_dataset){

	char* marshalled_dataset=NULL;
	int   enum_value=0;
	int   flag;

	/*recover marshalled_dataset: */
	marshalled_dataset=*pmarshalled_dataset;

	/*get enum value of ControlInput: */
	enum_value=ControlInputEnum;
	
	/*marshall enum: */
	memcpy(marshalled_dataset,&enum_value,sizeof(enum_value));marshalled_dataset+=sizeof(enum_value);
	
	/*marshall enum_type: */
	memcpy(marshalled_dataset,&enum_type,sizeof(enum_type));marshalled_dataset+=sizeof(enum_type);

	/*marshal values*/
	if(!values){
		flag=0;
		memcpy(marshalled_dataset,&flag,sizeof(flag));marshalled_dataset+=sizeof(flag);
	}
	else{
		flag=1;
		memcpy(marshalled_dataset,&flag,sizeof(flag));marshalled_dataset+=sizeof(flag);
		this->values->Marshall(&marshalled_dataset);
	}

	/*marshal savedvalues*/
	if(!savedvalues){
		flag=0;
		memcpy(marshalled_dataset,&flag,sizeof(flag));marshalled_dataset+=sizeof(flag);
	}
	else{
		flag=1;
		memcpy(marshalled_dataset,&flag,sizeof(flag));marshalled_dataset+=sizeof(flag);
		this->savedvalues->Marshall(&marshalled_dataset);
	}

	/*marshal gradient*/
	if(!gradient){
		flag=0;
		memcpy(marshalled_dataset,&flag,sizeof(flag));marshalled_dataset+=sizeof(flag);
	}
	else{
		flag=1;
		memcpy(marshalled_dataset,&flag,sizeof(flag));marshalled_dataset+=sizeof(flag);
		this->gradient->Marshall(&marshalled_dataset);
	}

	/*clean up and assign output pointer*/
	*pmarshalled_dataset=marshalled_dataset;
}
/*}}}*/
/*FUNCTION ControlInput::MarshallSize{{{1*/
int   ControlInput::MarshallSize(){
	
	int size=0;

	size=sizeof(enum_type)+
	  +3*sizeof(int) //3 flags
	  +sizeof(int); //sizeof(int) for enum value

	if(values)     size+=values->MarshallSize();
	if(savedvalues)size+=savedvalues->MarshallSize();
	if(gradient)   size+=gradient->MarshallSize();
	return size;
}
/*}}}*/
/*FUNCTION ControlInput::Demarshall{{{1*/
void  ControlInput::Demarshall(char** pmarshalled_dataset){

	char* marshalled_dataset=NULL;
	int   flag,input_enum_type;

	/*recover marshalled_dataset: */
	marshalled_dataset=*pmarshalled_dataset;

	/*this time, no need to get enum type, the pointer directly points to the beginning of the 
	 *object data (thanks to DataSet::Demarshall):*/
	memcpy(&enum_type,marshalled_dataset,sizeof(enum_type));marshalled_dataset+=sizeof(enum_type);

	/*Demarshal values*/
	memcpy(&flag,marshalled_dataset,sizeof(int));marshalled_dataset+=sizeof(int);
	if(flag){
		memcpy(&input_enum_type,marshalled_dataset,sizeof(int)); marshalled_dataset+=sizeof(int);
		if(input_enum_type==PentaVertexInputEnum){
			values=new PentaVertexInput();
			values->Demarshall(&marshalled_dataset);
		}
		else if(input_enum_type==TriaVertexInputEnum){
			values=new TriaVertexInput();
			values->Demarshall(&marshalled_dataset);
		}
		else ISSMERROR("Not supported yet");
	}
	else{
		values=NULL;
	}

	/*Demarshal savedvalues*/
	memcpy(&flag,marshalled_dataset,sizeof(int));marshalled_dataset+=sizeof(int);
	if(flag){
		memcpy(&input_enum_type,marshalled_dataset,sizeof(int)); marshalled_dataset+=sizeof(int);
		if(input_enum_type==PentaVertexInputEnum){
			savedvalues=new PentaVertexInput();
			savedvalues->Demarshall(&marshalled_dataset);
		}
		else if(input_enum_type==TriaVertexInputEnum){
			savedvalues=new TriaVertexInput();
			savedvalues->Demarshall(&marshalled_dataset);
		}
		else ISSMERROR("Not supported yet");
	}
	else{
		savedvalues=NULL;
	}

	/*Demarshal gradient*/
	memcpy(&flag,marshalled_dataset,sizeof(int));marshalled_dataset+=sizeof(int);
	if(flag){
		memcpy(&input_enum_type,marshalled_dataset,sizeof(int)); marshalled_dataset+=sizeof(int);
		if(input_enum_type==PentaVertexInputEnum){
			gradient=new PentaVertexInput();
			gradient->Demarshall(&marshalled_dataset);
		}
		else if(input_enum_type==TriaVertexInputEnum){
			gradient=new TriaVertexInput();
			gradient->Demarshall(&marshalled_dataset);
		}
		else ISSMERROR("Not supported yet");
	}
	else{
		gradient=NULL;
	}

	/*return: */
	*pmarshalled_dataset=marshalled_dataset;
	return;
}
/*}}}*/
/*FUNCTION ControlInput::Enum{{{1*/
int ControlInput::Enum(void){

	return ControlInputEnum;

}
/*}}}*/
/*FUNCTION ControlInput::copy{{{1*/
Object* ControlInput::copy() {
	
	ControlInput* output=NULL;

	output = new ControlInput();
	output->enum_type=this->enum_type;

	if(values)      output->values=(Input*)this->values->copy();
	if(savedvalues) output->savedvalues=(Input*)this->savedvalues->copy();
	if(gradient)    output->gradient=(Input*)this->gradient->copy();

	return output;
}
/*}}}*/
	
/*ControlInput management*/
/*FUNCTION ControlInput::EnumType{{{1*/
int ControlInput::EnumType(void){

	return this->enum_type;

}
/*}}}*/

/*Object functions*/
/*FUNCTION ControlInput::Constrain{{{1*/
void ControlInput::Constrain(double cm_min, double cm_max){
	values->Constrain(cm_min,cm_max);
}/*}}}*/
/*FUNCTION ControlInput::Extrude{{{1*/
void ControlInput::Extrude(void){
	values->Extrude();
	savedvalues->Extrude();
	//gradient->Extrude();
}/*}}}*/
/*FUNCTION ControlInput::GetGradient{{{1*/
void ControlInput::GetGradient(Vec gradient_vec,int* doflist){
	if(gradient) gradient->GetVectorFromInputs(gradient_vec,doflist);
}/*}}}*/
/*FUNCTION ControlInput::ScaleGradient{{{1*/
void ControlInput::ScaleGradient(double scaling_factor){
	if(!gradient) ISSMERROR("Gradient of ControlInput %s not found",EnumToString(enum_type));
	gradient->Scale(scaling_factor);
}/*}}}*/
/*FUNCTION ControlInput::SetGradient{{{1*/
void ControlInput::SetGradient(Input* gradient_in){
	if(gradient) delete gradient;
	gradient=gradient_in;
}/*}}}*/
/*FUNCTION ControlInput::SpawnResult{{{1*/
ElementResult* ControlInput::SpawnResult(int step, double time){
	return savedvalues->SpawnResult(step,time);
}/*}}}*/
/*FUNCTION ControlInput::SpawnTriaInput{{{1*/
Input* ControlInput::SpawnTriaInput(int* indices){
	return values->SpawnTriaInput(indices);
}/*}}}*/
/*FUNCTION ControlInput::SpawnGradient{{{1*/
ElementResult* ControlInput::SpawnGradient(int step, double time){
	gradient->SpawnResult(step,time);
}/*}}}*/
/*FUNCTION ControlInput::GetParameterAverage(double* pvalue){{{1*/
void ControlInput::GetParameterAverage(double* pvalue){
	values->GetParameterAverage(pvalue);
}/*}}}*/
/*FUNCTION ControlInput::GetParameterValue(bool* pvalue){{{1*/
void ControlInput::GetParameterValue(bool* pvalue){
	values->GetParameterValue(pvalue);
}/*}}}*/
/*FUNCTION ControlInput::GetParameterValue(int* pvalue){{{1*/
void ControlInput::GetParameterValue(int* pvalue){
	values->GetParameterValue(pvalue);
}/*}}}*/
/*FUNCTION ControlInput::GetParameterValue(double* pvalue){{{1*/
void ControlInput::GetParameterValue(double* pvalue){
	values->GetParameterValue(pvalue);
}/*}}}*/
/*FUNCTION ControlInput::GetParameterValue(double* pvalue){{{1*/
void ControlInput::GetParameterValue(double* pvalue,GaussTria* gauss){
	values->GetParameterValue(pvalue,gauss);
}/*}}}*/
/*FUNCTION ControlInput::GetParameterValue(double* pvalue){{{1*/
void ControlInput::GetParameterValue(double* pvalue,GaussPenta* gauss){
	values->GetParameterValue(pvalue,gauss);
}/*}}}*/
/*FUNCTION ControlInput::GetParameterDerivativeValue(double* derivativevalues, double* xyz_list, GaussTria* gauss){{{1*/
void ControlInput::GetParameterDerivativeValue(double* derivativevalues, double* xyz_list, GaussTria* gauss){
	values->GetParameterDerivativeValue(derivativevalues,xyz_list,gauss);
}/*}}}*/
/*FUNCTION ControlInput::GetParameterDerivativeValue(double* derivativevalues, double* xyz_list, GaussPenta* gauss){{{1*/
void ControlInput::GetParameterDerivativeValue(double* derivativevalues, double* xyz_list, GaussPenta* gauss){
	values->GetParameterDerivativeValue(derivativevalues,xyz_list,gauss);
}/*}}}*/
/*FUNCTION ControlInput::SaveValue{{{1*/
void ControlInput::SaveValue(void){
	if(!values) ISSMERROR("Values of %s not found",EnumToString(this->enum_type));

	if(savedvalues) delete this->savedvalues;
	this->savedvalues=(Input*)this->values->copy();
}/*}}}*/
/*FUNCTION ControlInput::UpdateValue{{{1*/
void ControlInput::UpdateValue(double scalar){
	if(!gradient)    ISSMERROR("Gradient of %s not found",EnumToString(this->enum_type));
	if(!savedvalues) ISSMERROR("Values of %s not found",EnumToString(this->enum_type));

	if(values) delete this->values;
	this->values=(Input*)this->savedvalues->copy();
	this->values->AXPY(gradient,scalar);
}/*}}}*/
/*FUNCTION ControlInput::VerticallyIntegrate{{{1*/
void ControlInput::VerticallyIntegrate(Input* thickness_input){
	values->VerticallyIntegrate(thickness_input);
}/*}}}*/
